r"""Contain an analyzer that processes a mapping of analyzers."""

from __future__ import annotations

__all__ = ["MappingAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from arkas.analyzer.base import BaseAnalyzer
from arkas.output.mapping import OutputDict

if TYPE_CHECKING:
    from collections.abc import Mapping

    import polars as pl


logger = logging.getLogger(__name__)


class MappingAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that processes a mapping of analyzers.

    Args:
        analyzers: The mapping of analyzers.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import (
    ...     MappingAnalyzer,
    ...     AccuracyAnalyzer,
    ...     BalancedAccuracyAnalyzer,
    ... )
    >>> analyzer = MappingAnalyzer(
    ...     {
    ...         "one": AccuracyAnalyzer(y_true="target", y_pred="pred"),
    ...         "two": BalancedAccuracyAnalyzer(y_true="target", y_pred="pred"),
    ...     }
    ... )
    >>> analyzer
    MappingAnalyzer(
      (one): AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
      (two): BalancedAccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
    )
    >>> frame = pl.DataFrame({"pred": [1, 0, 0, 1, 1], "target": [1, 0, 0, 1, 1]})
    >>> output = analyzer.analyze(frame)
    >>> output
    OutputDict(count=2)

    ```
    """

    def __init__(self, analyzers: Mapping[str, BaseAnalyzer]) -> None:
        self._analyzers = analyzers

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._analyzers))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> OutputDict:
        return OutputDict(
            {
                key: analyzer.analyze(frame=frame, lazy=lazy)
                for key, analyzer in self._analyzers.items()
            }
        )



r"""Contain data utility functions."""

from __future__ import annotations

__all__ = ["find_keys", "find_missing_keys", "flat_keys", "prepare_array"]


from typing import TYPE_CHECKING

import numpy as np
import polars as pl

from arkas.utils.array import to_array

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


def find_keys(data: Mapping | pl.DataFrame) -> set:
    r"""Find all the keys in the input data.

    Args:
        data: The input data.

    Returns:
        The set of keys.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.utils.data import find_keys
    >>> keys = find_keys(
    ...     {"pred": np.array([3, 2, 0, 1, 0]), "target": np.array([3, 2, 0, 1, 0])}
    ... )
    >>> sorted(keys)
    ['pred', 'target']
    >>> keys = find_keys(pl.DataFrame({"pred": [3, 2, 0, 1, 0], "target": [3, 2, 0, 1, 0]}))
    >>> sorted(keys)
    ['pred', 'target']

    ```
    """
    if isinstance(data, pl.DataFrame):
        return set(data.columns)
    return set(data.keys())


def find_missing_keys(keys: set | Sequence, queries: set | Sequence) -> set:
    r"""Return the set of queries that are not in the input keys.

    Args:
        keys: The keys.
        queries: The queries i.e. the keys to check in the input keys.

    Returns:
        The set of missing keys.

    Example usage:

    ```pycon

    >>> from arkas.utils.data import find_missing_keys
    >>> keys = find_missing_keys(
    ...     keys={"key1", "key2", "key3"}, queries=["key1", "key2", "key4"]
    ... )
    >>> keys
    {'key4'}

    ```
    """
    keys = set(keys)
    queries = set(queries)
    intersection = set(keys).intersection(queries)
    return queries.difference(intersection)


def flat_keys(keys: Sequence[str | Sequence[str]]) -> list[str]:
    r"""Flat and merge a sequence of keys or sequence of keys.

    Args:
        keys: The keys to flat.

    Returns:
        The list of keys.

    Example usage:

    ```pycon

    >>> from arkas.utils.data import flat_keys
    >>> keys = flat_keys(["key0", ["key1"], ["key2", "key3", "key4"]])
    >>> keys
    ['key0', 'key1', 'key2', 'key3', 'key4']

    ```
    """
    out = []
    for key in keys:
        if isinstance(key, str):
            out.append(key)
        else:
            out.extend(key)
    return out


def prepare_array(data: dict | pl.DataFrame, keys: Sequence[str]) -> np.ndarray:
    r"""Prepare the data and return an array.

    Args:
        data: The input data.
        keys: The keys or columns to extract from the input data.

    Returns:
        The prepared array.

    Example usage:

    ```pycon

    >>> from arkas.utils.data import prepare_array
    >>> keys = prepare_array({"key": np.array([1, 2, 3, 4, 5])}, keys="key")
    >>> keys
    array([1, 2, 3, 4, 5])

    ```
    """
    if isinstance(keys, str):
        return to_array(data[keys])
    if isinstance(data, pl.DataFrame):
        return to_array(data.select(keys))
    return np.stack([to_array(data[key]) for key in keys], axis=1)



r"""Contain a simple runner to analyze data."""

from __future__ import annotations

__all__ = ["AnalysisRunner"]

import logging

from coola.utils import str_indent, str_mapping
from grizz.ingestor import BaseIngestor, setup_ingestor
from grizz.transformer import BaseTransformer, setup_transformer
from iden.utils.time import timeblock

from arkas.analyzer.base import BaseAnalyzer, setup_analyzer
from arkas.exporter import BaseExporter, setup_exporter
from arkas.runner.base import BaseRunner

logger = logging.getLogger(__name__)


class AnalysisRunner(BaseRunner):
    r"""Implement a runner to analyze data.

    Args:
        ingestor: The data ingestor or its configuration.
        transformer: The data transformer or its configuration.
        analyzer: The analyzer or its configuration.
        exporter: The output exporter or its configuration.
        lazy: If ``True``, the analyzer computation is done lazily.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> import polars as pl
    >>> from pathlib import Path
    >>> from grizz.ingestor import Ingestor
    >>> from grizz.transformer import SequentialTransformer
    >>> from arkas.analyzer import AccuracyAnalyzer
    >>> from arkas.exporter import MetricExporter
    >>> from arkas.runner import AnalysisRunner
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     runner = AnalysisRunner(
    ...         ingestor=Ingestor(
    ...             pl.DataFrame(
    ...                 {
    ...                     "pred": [3, 2, 0, 1, 0],
    ...                     "target": [3, 2, 0, 1, 0],
    ...                 }
    ...             )
    ...         ),
    ...         transformer=SequentialTransformer(transformers=[]),
    ...         analyzer=AccuracyAnalyzer(y_true="target", y_pred="pred"),
    ...         exporter=MetricExporter(Path(tmpdir).joinpath("metrics.pkl")),
    ...     )
    ...     print(runner)
    ...     runner.run()
    ...
    AnalysisRunner(
      (ingestor): Ingestor(shape=(5, 2))
      (transformer): SequentialTransformer()
      (analyzer): AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
      (exporter): MetricExporter(
          (path): .../metrics.pkl
          (saver): PickleSaver(protocol=5)
          (exist_ok): False
          (show_metrics): True
        )
      (lazy): True
    )

    ```
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        transformer: BaseTransformer | dict,
        analyzer: BaseAnalyzer | dict,
        exporter: BaseExporter | dict,
        lazy: bool = True,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._transformer = setup_transformer(transformer)
        self._analyzer = setup_analyzer(analyzer)
        self._exporter = setup_exporter(exporter)
        self._lazy = lazy

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "transformer": self._transformer,
                    "analyzer": self._analyzer,
                    "exporter": self._exporter,
                    "lazy": self._lazy,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def run(self) -> None:
        with timeblock():
            self._run()

    def _run(self) -> None:
        logger.info("Ingesting data...")
        raw_data = self._ingestor.ingest()
        logger.info("Transforming data...")
        data = self._transformer.transform(raw_data)
        logger.info("Analyzing...")
        output = self._analyzer.analyze(data, lazy=self._lazy)
        logger.info(f"output:\n{output}")
        logger.info("Exporting the output...")
        self._exporter.export(output)



r"""Contain scale utility functions."""

from __future__ import annotations

__all__ = ["auto_yscale_continuous", "auto_yscale_discrete"]


import numpy as np

from arkas.utils.array import nonnan


def auto_yscale_continuous(array: np.ndarray, nbins: int | None = None) -> str:
    r"""Find a good scale for y-axis based on the data distribution.

    Args:
        array: The data to use to find the scale.
        nbins: The number of bins in the histogram.

    Returns:
        The scale for y-axis.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.plot.utils import auto_yscale_continuous
    >>> auto_yscale_continuous(np.arange(100))
    linear

    ```
    """
    if nbins is None:
        nbins = 100
    array = nonnan(array)
    counts = np.histogram(array, bins=nbins)[0]
    nonzero_count = [c for c in counts if c > 0]
    if len(nonzero_count) <= 2 or (max(nonzero_count) / max(min(nonzero_count), 1)) < 50:
        return "linear"
    if np.nanmin(array) <= 0.0:
        return "symlog"
    return "log"


def auto_yscale_discrete(min_count: int, max_count: int, threshold: int = 50) -> str:
    r"""Find a good scale for y-axis based on the data distribution.

    Args:
        min_count: The minimal count value.
        max_count: The maximal count value.
        threshold: The threshold used to control the transition from
            linear to log scale. ``50`` means the ratio
            ``max_count/min_count`` must be greater than 50 to use the
            log scale. If it is lower than 50, the linear scale is
            used.

    Returns:
        The scale for y-axis.

    Example usage:

    ```pycon

    >>> from arkas.plot.utils import auto_yscale_discrete
    >>> auto_yscale_discrete(min_count=5, max_count=10)
    linear
    >>> auto_yscale_discrete(min_count=5, max_count=1000)
    log

    ```
    """
    return "log" if (max_count / max(min_count, 1)) >= threshold else "linear"



r"""Implement a result that merges multiple results."""

from __future__ import annotations

__all__ = ["SequentialResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import str_indent, str_sequence

from arkas.result import BaseResult

if TYPE_CHECKING:
    from collections.abc import Sequence


class SequentialResult(BaseResult):
    r"""Implement a result to merge multiple result objects into a single
    result object.

    Args:
        results: The results to merge. This order is used to merge
            the metrics and figures if they have duplicate keys,
            i.e. only the last value for each key is kept.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import SequentialResult, Result
    >>> result = SequentialResult(
    ...     [
    ...         Result(metrics={"accuracy": 62.0, "count": 42}),
    ...         Result(metrics={"ap": 0.42, "count": 42}),
    ...     ]
    ... )
    >>> result
    SequentialResult(count=2)
    >>> result.compute_metrics()
    {'accuracy': 62.0, 'count': 42, 'ap': 0.42}

    ```
    """

    def __init__(self, results: Sequence[BaseResult]) -> None:
        self._results = tuple(results)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={len(self._results):,})"

    def __str__(self) -> str:
        args = str_indent(str_sequence(self._results))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict:
        out = {}
        for result in self._results:
            out |= result.compute_metrics(prefix=prefix, suffix=suffix)
        return out

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._results, other._results, equal_nan=equal_nan)

    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict:
        out = {}
        for result in self._results:
            out |= result.generate_figures(prefix=prefix, suffix=suffix)
        return out



r"""Implement the Jensen-Shannon (JS) divergence between two
distributions."""

from __future__ import annotations

__all__ = ["jensen_shannon_divergence"]


from typing import TYPE_CHECKING

from arkas.metric.distribution.kl import _kl_divergence
from arkas.metric.utils import preprocess_same_shape_arrays
from arkas.utils.imports import check_scipy, is_scipy_available

if is_scipy_available():
    pass

if TYPE_CHECKING:
    import numpy as np


def jensen_shannon_divergence(
    p: np.ndarray,
    q: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
) -> dict[str, float]:
    r"""Return the Jensen-Shannon (JS) divergence between two
    distributions.

    Args:
        p: The true probability distribution.
        q: The model probability distribution.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import jensen_shannon_divergence
    >>> jensen_shannon_divergence(
    ...     p=np.array([0.1, 0.6, 0.1, 0.2]), q=np.array([0.2, 0.5, 0.2, 0.1])
    ... )
    {'size': 4, 'jensen_shannon_divergence': 0.027...}

    ```
    """
    check_scipy()
    p, q = preprocess_same_shape_arrays(arrays=[p.ravel(), q.ravel()])

    size = p.size
    div = float("nan")
    if size > 0:
        m = 0.5 * (p + q)
        div = 0.5 * (_kl_divergence(p, m) + _kl_divergence(q, m))
    return {f"{prefix}size{suffix}": size, f"{prefix}jensen_shannon_divergence{suffix}": div}



r"""Contain some functions to validate variables."""

from __future__ import annotations

__all__ = ["check_positive"]


def check_positive(name: str, value: float) -> None:
    r"""Check the value is positive (>=0).

    Args:
        name: The name of the variable.
        value: The value to check.

    Raises:
        ValueError: if the value is not positive.

    Example usage:

    ```pycon

    >>> from arkas.utils.validation import check_positive
    >>> check_positive("var", 1)

    ```
    """
    if value < 0:
        msg = f"Incorrect {name!r}: {value}. The value must be positive"
        raise ValueError(msg)



r"""Contain the implementation of a HTML content generator that analyzes
accuracy performances."""

from __future__ import annotations

__all__ = ["BalancedAccuracyContentGenerator", "create_template"]

import logging
from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping
from jinja2 import Template

from arkas.content.section import BaseSectionContentGenerator
from arkas.evaluator2.balanced_accuracy import BalancedAccuracyEvaluator
from arkas.metric.utils import check_nan_policy

if TYPE_CHECKING:
    from arkas.state import AccuracyState


logger = logging.getLogger(__name__)


class BalancedAccuracyContentGenerator(BaseSectionContentGenerator):
    r"""Implement a HTML content generator that analyzes balanced
    accuracy performances.

    Args:
        state: The data structure containing the states.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.content import BalancedAccuracyContentGenerator
    >>> from arkas.state import AccuracyState
    >>> generator = BalancedAccuracyContentGenerator(
    ...     state=AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> generator
    BalancedAccuracyContentGenerator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )

    ```
    """

    def __init__(self, state: AccuracyState, nan_policy: str = "propagate") -> None:
        self._state = state

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"state": self._state, "nan_policy": self._nan_policy}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(str_mapping({"state": self._state, "nan_policy": self._nan_policy}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._state.equal(other._state, equal_nan=equal_nan)
            and self._nan_policy == other._nan_policy
        )

    def generate_content(self) -> str:
        logger.info("Generating the balance accuracy content...")
        metrics = BalancedAccuracyEvaluator(self._state, nan_policy=self._nan_policy).evaluate()
        return Template(create_template()).render(
            {
                "balanced_accuracy": f"{metrics.get('balanced_accuracy', float('nan')):.4f}",
                "count": f"{metrics.get('count', 0):,}",
                "y_true_name": self._state.y_true_name,
                "y_pred_name": self._state.y_pred_name,
            }
        )


def create_template() -> str:
    r"""Return the template of the content.

    Returns:
        The content template.

    Example usage:

    ```pycon

    >>> from arkas.content.accuracy import create_template
    >>> template = create_template()

    ```
    """
    return """<ul>
  <li><b>balanced accuracy</b>: {{balanced_accuracy}}</li>
  <li><b>number of samples</b>: {{count}}</li>
  <li><b>target label column</b>: {{y_true_name}}</li>
  <li><b>predicted label column</b>: {{y_pred_name}}</li>
</ul>
"""



r"""Contain data plotters."""

from __future__ import annotations

__all__ = ["BasePlotter", "ColumnCooccurrencePlotter", "Plotter", "PlotterDict"]

from arkas.plotter.base import BasePlotter
from arkas.plotter.column_cooccurrence import ColumnCooccurrencePlotter
from arkas.plotter.mapping import PlotterDict
from arkas.plotter.vanilla import Plotter



r"""Implement the balanced accuracy evaluator."""

from __future__ import annotations

__all__ = ["BalancedAccuracyEvaluator"]


from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping

from arkas.evaluator2.base import BaseEvaluator
from arkas.evaluator2.vanilla import Evaluator
from arkas.metric import balanced_accuracy
from arkas.metric.utils import check_nan_policy

if TYPE_CHECKING:
    from arkas.state.accuracy import AccuracyState


class BalancedAccuracyEvaluator(BaseEvaluator):
    r"""Implement the balanced accuracy evaluator.

    Args:
        state: The state containing the ground truth and predicted
            labels.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.evaluator2 import BalancedAccuracyEvaluator
    >>> from arkas.state import AccuracyState
    >>> evaluator = BalancedAccuracyEvaluator(
    ...     AccuracyState(
    ...         y_true=np.array([1, 0, 0, 1, 1]),
    ...         y_pred=np.array([1, 0, 0, 1, 1]),
    ...         y_true_name="target",
    ...         y_pred_name="pred",
    ...     )
    ... )
    >>> evaluator
    BalancedAccuracyEvaluator(
      (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
      (nan_policy): propagate
    )
    >>> evaluator.evaluate()
    {'balanced_accuracy': 1.0, 'count': 5}

    ```
    """

    def __init__(self, state: AccuracyState, nan_policy: str = "propagate") -> None:
        self._state = state
        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"state": self._state, "nan_policy": self._nan_policy}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> Evaluator:
        return Evaluator(metrics=self.evaluate())

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._state.equal(other._state, equal_nan=equal_nan)
            and self._nan_policy == other._nan_policy
        )

    def evaluate(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return balanced_accuracy(
            y_true=self._state.y_true,
            y_pred=self._state.y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )



r"""Implement a simple output."""

from __future__ import annotations

__all__ = ["Output"]

from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping

from arkas.output.lazy import BaseLazyOutput

if TYPE_CHECKING:
    from arkas.content.base import BaseContentGenerator
    from arkas.evaluator2.base import BaseEvaluator
    from arkas.plotter.base import BasePlotter


class Output(BaseLazyOutput):
    r"""Implement a simple output.

    Args:
        content: The HTML content generator.
        evaluator: The evaluator.
        plotter: The plotter.

    Example usage:

    ```pycon

    >>> from arkas.output import Output
    >>> from arkas.content import ContentGenerator
    >>> from arkas.evaluator2 import Evaluator
    >>> from arkas.plotter import Plotter
    >>> output = Output(
    ...     content=ContentGenerator("meow"), evaluator=Evaluator(), plotter=Plotter()
    ... )
    >>> output
    Output(
      (content): ContentGenerator()
      (evaluator): Evaluator(count=0)
      (plotter): Plotter(count=0)
    )
    >>> output.get_content_generator()
    ContentGenerator()
    >>> output.get_evaluator()
    Evaluator(count=0)
    >>> output.get_plotter()
    Plotter(count=0)

    ```
    """

    def __init__(
        self,
        content: BaseContentGenerator,
        evaluator: BaseEvaluator,
        plotter: BasePlotter,
    ) -> None:
        self._content = content
        self._evaluator = evaluator
        self._plotter = plotter

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "content": self._content,
                    "evaluator": self._evaluator,
                    "plotter": self._plotter,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self._content.equal(other._content, equal_nan=equal_nan)
            and self._evaluator.equal(other._evaluator, equal_nan=equal_nan)
            and self._plotter.equal(other._plotter, equal_nan=equal_nan)
        )

    def _get_content_generator(self) -> BaseContentGenerator:
        return self._content

    def _get_evaluator(self) -> BaseEvaluator:
        return self._evaluator

    def _get_plotter(self) -> BasePlotter:
        return self._plotter



r"""Implement a pairwise column co-occurrence analyzer."""

from __future__ import annotations

__all__ = ["ColumnCooccurrenceAnalyzer"]

import logging
from typing import TYPE_CHECKING

from grizz.utils.format import str_shape_diff

from arkas.analyzer.lazy import BaseInNLazyAnalyzer
from arkas.output.column_cooccurrence import ColumnCooccurrenceOutput

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from arkas.figure import BaseFigureConfig

logger = logging.getLogger(__name__)


class ColumnCooccurrenceAnalyzer(BaseInNLazyAnalyzer):
    r"""Implement a pairwise column co-occurrence analyzer.

    Args:
        columns: The columns to analyze. If ``None``, it analyzes all
            the columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        ignore_self: If ``True``, the diagonal of the co-occurrence
            matrix (a.k.a. self-co-occurrence) is set to 0.
        figure_config: The figure configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import ColumnCooccurrenceAnalyzer
    >>> analyzer = ColumnCooccurrenceAnalyzer()
    >>> analyzer
    ColumnCooccurrenceAnalyzer(columns=None, exclude_columns=(), missing_policy='raise', ignore_self=False, figure_config=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> output = analyzer.analyze(frame)
    >>> output
    ColumnCooccurrenceOutput(shape=(7, 3), ignore_self=False)

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        ignore_self: bool = False,
        figure_config: BaseFigureConfig | None = None,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._ignore_self = ignore_self
        self._figure_config = figure_config

    def get_args(self) -> dict:
        return super().get_args() | {
            "ignore_self": self._ignore_self,
            "figure_config": self._figure_config,
        }

    def _analyze(self, frame: pl.DataFrame) -> ColumnCooccurrenceOutput:
        logger.info(
            "Analyzing the pairwise column co-occurrence of "
            f"{len(self.find_columns(frame)):,}..."
        )
        columns = self.find_common_columns(frame)
        out = frame.select(columns)
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return ColumnCooccurrenceOutput(
            frame=out, ignore_self=self._ignore_self, figure_config=self._figure_config
        )
